unit Generators;

{ Johan hlgren 2014 }

interface

uses
  SysUtils, Generics.Collections;

type
  TMapFun<T> = reference to function (Arg1: T): T;
  TStrFun<T> = reference to function (Arg1: T): String;
  TFoldFun<T> = reference to function(Arg1, Arg2: T): T;
  TPredicate<T> = reference to function (Arg1: T): Boolean;
  TNextItemFunc<T> = reference to function (out Item: T): Boolean;

  // Generic general-purpose generators ----------

  IGenerator<T> = interface
    function GetNextItem(out Item: T): Boolean;
    function Filter(Pred: TPredicate<T>): IGenerator<T>;
    function Take(Num: Integer): IGenerator<T>;
    function Map(MapFun: TMapFun<T>): IGenerator<T>;
    function TakeWhile(Pred: TPredicate<T>): IGenerator<T>;
    function DropWhile(Pred: TPredicate<T>): IGenerator<T>;
    function ToList: TList<T>;
    function ToStr(StrFun: TStrFun<T>): String;
    function Fold(FoldFun: TFoldFun<T>; InitialValue: T): T;
    procedure Kickoff;
  end;

  TGenerator<T> = class(TInterfacedObject, IGenerator<T>)
  protected
    NextGetNextItem: TNextItemFunc<T>;
  public
    function GetNextItem(out Item: T): Boolean; virtual;
  public
    function Filter(Pred: TPredicate<T>): IGenerator<T>;
    function Take(Num: Integer): IGenerator<T>;
    function Map(MapFun: TMapFun<T>): IGenerator<T>;
    function TakeWhile(Pred: TPredicate<T>): IGenerator<T>;
    function DropWhile(Pred: TPredicate<T>): IGenerator<T>;

    function ToList: TList<T>;
    function ToStr(StrFun: TStrFun<T>): String;
    function Fold(FoldFun: TFoldFun<T>; InitialValue: T): T;
    procedure Kickoff;
    constructor Create(NextGetNextItem: TNextItemFunc<T>);
  end;

  TFilterGenerator<T> = class(TGenerator<T>)
  protected
    Predicate: TPredicate<T>;
  public
    function GetNextItem(out Item: T): Boolean; override;
    constructor Create(NextGetNextItem: TNextItemFunc<T>; Predicate: TPredicate<T>);
  end;

  TTakeGenerator<T> = class(TGenerator<T>)
  protected
    TakenSoFar: Integer;
    ItemsToTake: Integer;
  public
    function GetNextItem(out Item: T): Boolean; override;
    constructor Create(NextGetNextItem: TNextItemFunc<T>; ItemsToTake: Integer);
  end;

  TMapGenerator<T> = class(TGenerator<T>)
  protected
    MapFun: TMapFun<T>;
  public
    function GetNextItem(out Item: T): Boolean; override;
    constructor Create(NextGetNextItem: TNextItemFunc<T>; MapFun: TMapFun<T>);
  end;

  TTakeWhileGenerator<T> = class(TGenerator<T>)
  protected
    Predicate: TPredicate<T>;
  public
    function GetNextItem(out Item: T): Boolean; override;
    constructor Create(NextGetNextItem: TNextItemFunc<T>; Predicate: TPredicate<T>);
  end;

  TDropWhileGenerator<T> = class(TGenerator<T>)
  protected
    Predicate: TPredicate<T>;
  public
    function GetNextItem(out Item: T): Boolean; override;
    constructor Create(NextGetNextItem: TNextItemFunc<T>; Predicate: TPredicate<T>);
  end;

  TListGenerator<T> = class(TGenerator<T>)
  protected
    Index: Integer;
    List: TList<T>;
  public
    function GetNextItem(out Item: T): Boolean; override;
    constructor Create(List: TList<T>);
  end;

  // ---------- Specific generators ----------

  TIntegersGenerator = class(TGenerator<Integer>)
  protected
    CurrentValue, Step, MaxValue: Integer;
  public
    function GetNextItem(out Item: Integer): Boolean; override;
    constructor Create(StartValue: Integer = 1; Step: Integer = 1; MaxValue: Integer = MAXINT);
  end;

  TFilesGenerator = class(TGenerator<String>)
  protected
    Files, Dirs: TStack<String>;
    procedure Process(const Dir: String);
  public
    function GetNextItem(out Item: String): Boolean; override;
    constructor Create(const BaseDirectory: String); overload;
    destructor Destroy; override;
  end;

  // -----------------------------------------------------------------------------------------------

  (* TODO REMOVE
  IConversionGenerator<T_old,T_new> = interface
    function Get: IGenerator<T_new>;
    function GetNextItem(out Item: T_new): Boolean;
  end;

  TConversionGenerator<T_old,T_new> = class(TInterfacedObject, IConversionGenerator<T_old,T_new>)
  protected
    NextGenerator: IGenerator<T_Old>;
    Converter: TFunc<T_old,T_new>;
  public
    function Get: IGenerator<T_new>;
    function GetNextItem(out Item: T_new): Boolean;
    constructor Create(NextGenerator: IGenerator<T_old>; Converter: TFunc<T_old,T_new>);
  end;
  *)

// Convenience functions

function Ints(StartValue: Integer = 1; Step: Integer = 1;
  MaxValue: Integer = MAXINT): IGenerator<Integer>;

function FilesGen(const BaseDirectory: String): IGenerator<String>;

implementation

{ TGenerator<T> }

function TGenerator<T>.GetNextItem(out Item: T): Boolean;
begin
  Result := False;
end;

function TGenerator<T>.Filter(Pred: TPredicate<T>): IGenerator<T>;
begin
  Result := TFilterGenerator<T>.Create(Self.GetNextItem, Pred);
end;

function TGenerator<T>.Map(MapFun: TMapFun<T>): IGenerator<T>;
begin
  Result := TMapGenerator<T>.Create(Self.GetNextItem, MapFun);
end;

function TGenerator<T>.Take(Num: Integer): IGenerator<T>;
begin
  Result := TTakeGenerator<T>.Create(Self.GetNextItem, Num);
end;

function TGenerator<T>.TakeWhile(Pred: TPredicate<T>): IGenerator<T>;
begin
  Result := TTakeWhileGenerator<T>.Create(Self.GetNextItem, Pred);
end;

function TGenerator<T>.DropWhile(Pred: TPredicate<T>): IGenerator<T>;
begin
  Result := TDropWhileGenerator<T>.Create(Self.GetNextItem, Pred);
end;

function TGenerator<T>.ToList: TList<T>;
var
  Item: T;
begin
  Result := TList<T>.Create;
  while GetNextItem(Item) do
    Result.Add(Item);
end;

function TGenerator<T>.Fold(FoldFun: TFoldFun<T>; InitialValue: T): T;
var
  Item: T;
begin
  Result := InitialValue;
  while GetNextItem(Item) do
    Result := FoldFun(Item, Result);
end;

function TGenerator<T>.ToStr(StrFun: TStrFun<T>): String;
var
  Item: T;
begin
  Result := '[';
  while GetNextItem(Item) do
    Result := Result + StrFun(Item) + ',';
  if Length(Result) > 1 then
    Result := Copy(Result, 1, Length(Result)-1);
  Result := Result + ']';
end;

procedure TGenerator<T>.Kickoff;
var
  Item: T;
begin
  while GetNextItem(Item) do
    ;
end;

constructor TGenerator<T>.Create(NextGetNextItem: TNextItemFunc<T>);
begin
  Self.NextGetNextItem := NextGetNextItem;
end;

{ TFilterGenerator<T> }

function TFilterGenerator<T>.GetNextItem(out Item: T): Boolean;
begin
  while NextGetNextItem(Item) do
    if Predicate(Item) then
      Exit(True);
  Exit(False);
end;

constructor TFilterGenerator<T>.Create(NextGetNextItem: TNextItemFunc<T>; Predicate: TPredicate<T>);
begin
  inherited Create(NextGetNextItem);
  Self.Predicate := Predicate;
end;

{ TTakeGenerator<T> }

function TTakeGenerator<T>.GetNextItem(out Item: T): Boolean;
begin
  if TakenSoFar < ItemsToTake then
  begin
    Result := NextGetNextItem(Item);
    Inc(TakenSoFar);
  end
  else
    Result := False;
end;

constructor TTakeGenerator<T>.Create(NextGetNextItem: TNextItemFunc<T>; ItemsToTake: Integer);
begin
  inherited Create(NextGetNextItem);
  Self.TakenSoFar := 0;
  Self.ItemsToTake := ItemsToTake;
end;

{ TMapGenerator<T> }

function TMapGenerator<T>.GetNextItem(out Item: T): Boolean;
begin
  Result := NextGetNextItem(Item);
  if Result then
    Item := MapFun(Item);
end;

constructor TMapGenerator<T>.Create(NextGetNextItem: TNextItemFunc<T>; MapFun: TMapFun<T>);
begin
  inherited Create(NextGetNextItem);
  Self.MapFun := MapFun;
end;

{ TTakeWhileGenerator<T> }

function TTakeWhileGenerator<T>.GetNextItem(out Item: T): Boolean;
begin
  Result := NextGetNextItem(Item) and Predicate(Item);
end;

constructor TTakeWhileGenerator<T>.Create(NextGetNextItem: TNextItemFunc<T>; Predicate: TPredicate<T>);
begin
  inherited Create(NextGetNextItem);
  Self.Predicate := Predicate;
end;

{ TDropWhileGenerator<T> }

function TDropWhileGenerator<T>.GetNextItem(out Item: T): Boolean;
var
  ResultOfPred: Boolean;
begin
  repeat
    Result := NextGetNextItem(Item);
    if Result then
      ResultOfPred := Predicate(Item);
  until Result and not ResultOfPred;
end;

constructor TDropWhileGenerator<T>.Create(NextGetNextItem: TNextItemFunc<T>; Predicate: TPredicate<T>);
begin
  inherited Create(NextGetNextItem);
  Self.Predicate := Predicate;
end;

{ TListGenerator<T> }

function TListGenerator<T>.GetNextItem(out Item: T): Boolean;
begin
  if Index < List.Count then
  begin
    Item := List[Index];
    Inc(Index);
    Result := True;
  end
  else
    Result := False;
end;

constructor TListGenerator<T>.Create(List: TList<T>);
begin
  Self.List := List;
  Index := 0;
end;

{ TIntegersGenerator }

function TIntegersGenerator.GetNextItem(out Item: Integer): Boolean;
begin
  Result := CurrentValue <= MaxValue;
  if Result then
  begin
    Item := CurrentValue;
    Inc(CurrentValue, Step);
  end;
end;

constructor TIntegersGenerator.Create(StartValue: Integer = 1; Step: Integer = 1;
  MaxValue: Integer = MAXINT);
begin
  Self.Step := Step;
  Self.MaxValue := MaxValue;
  Self.CurrentValue := StartValue;
end;

{ TFilesGenerator }

procedure TFilesGenerator.Process(const Dir: String);
var
  SR: TSearchRec;
begin
  if FindFirst(Dir + '*.*', faAnyFile, SR) = 0 then
  begin
    repeat
      if SR.Name[1] <> '.' then
      begin
        if (SR.Attr and faDirectory) = faDirectory then
          Dirs.Push(IncludeTrailingPathDelimiter(Dir + SR.Name))
        else
          Files.Push(Dir + SR.Name);
      end;
    until FindNext(SR) <> 0;
    FindClose(SR);
  end;
end;

function TFilesGenerator.GetNextItem(out Item: String): Boolean;
begin
  if Files.Count > 0 then
  begin
    Item := Files.Pop;
    Exit(True);
  end;

  while Dirs.Count > 0 do
  begin
    Process(Dirs.Pop);
    if Files.Count > 0 then
    begin
      Item := Files.Pop;
      Exit(True);
    end;
  end;

  Exit(False);
end;

constructor TFilesGenerator.Create(const BaseDirectory: String);
begin
  Dirs := TStack<String>.Create;
  Files := TStack<String>.Create;
  Process(IncludeTrailingPathDelimiter(BaseDirectory));
end;

destructor TFilesGenerator.Destroy;
begin
  Files.Free;
  Dirs.Free;
  inherited;
end;

function Ints(StartValue: Integer = 1; Step: Integer = 1;
  MaxValue: Integer = MAXINT): IGenerator<Integer>;
begin
  Result := TIntegersGenerator.Create(StartValue, Step, MaxValue);
end;

function FilesGen(const BaseDirectory: String): IGenerator<String>;
begin
  Result := TFilesGenerator.Create(BaseDirectory);
end;

(* TODO REMOVE
{ TConversionGenerator<T_old, T_new> }

constructor TConversionGenerator<T_old, T_new>.Create(NextGenerator: IGenerator<T_old>;
  Converter: TFunc<T_old, T_new>);
begin
  Self.NextGenerator := NextGenerator;
  Self.Converter := Converter;
end;

function TConversionGenerator<T_old, T_new>.Get: IGenerator<T_new>;
begin
  Result := TTakeGenerator<T_new>.Create(Self.GetNextItem, MAXINT); // todo ID-generator???
end;

function TConversionGenerator<T_old, T_new>.GetNextItem(out Item: T_new): Boolean;
var
  OldItem: T_old;
begin
  Result := NextGenerator.GetNextItem(OldItem);
  if Result then
    Item := Converter(OldItem);
end;
*)

end.
